from __future__ import annotations

import logging
import os
import pickle
import inspect
import types
from pathlib import Path
from typing import Type, Union, Dict, Tuple, Optional, TypeVar, Callable, List

import networkx as nx
import requests
from tqdm import tqdm

from wikes_toolkit.esbm.esbm_eval import ESBMSummaryEvaluator
from wikes_toolkit.esbm.esbm_graph_components import ESBMBaseGraph
from wikes_toolkit.esbm.esbm_pandas_graph import PandasESBMGraph
from wikes_toolkit.wikes.wikes_eval import WikESSummaryEvaluator
from wikes_toolkit.wikes.wikes_versions import WikESVersions
from wikes_toolkit.esbm.esbm_versions import ESBMVersions

from wikes_toolkit.base.graph_components import BaseESGraph, RootEntity, Triple

from wikes_toolkit.esbm.esbm_graph import ESBMGraph
from wikes_toolkit.wikes.wikes_graph import WikESGraph
from wikes_toolkit.wikes.wikes_pandas_graph import PandasWikESGraph
from wikes_toolkit.base.versions import DatasetName, DatasetVersion
from wikes_toolkit.wikes.wikies_graph_components import WikESBaseGraph

logger = logging.getLogger(__name__)


class WikESToolkit:
    T = TypeVar('T', bound=BaseESGraph)
    WikES_datasets = WikESVersions.available_versions()
    ESBM_datasets = ESBMVersions.available_versions()

    def __init__(self, save_path: Union[str, Path] = None, log_level: int = logging.DEBUG):
        if save_path is None:
            save_path = Path.home() / '.wikes_data'
        self.save_path = Path(save_path)
        self.save_path.mkdir(parents=True, exist_ok=True)
        if not self.save_path.exists():
            raise Exception("WikES could not initialize the save path...")
        logging.basicConfig(level=log_level)

    def __download_graph(self, dataset: DatasetName) -> None:
        url = dataset.get_dataset_url()
        response = requests.get(url, stream=True)
        if response.status_code != 200:
            raise Exception(
                f"Failed to download dataset [{dataset}] from {url}. HTTP Status Code: {response.status_code}")

        total_size = int(response.headers.get('content-length', 0))
        dataset_path = self.save_path / dataset.get_version() / f"{dataset.value}.pkl"
        dataset_path.parent.mkdir(parents=True, exist_ok=True)
        with open(dataset_path, 'wb') as file, tqdm(
                desc=str(dataset_path),
                total=total_size,
                unit='iB',
                unit_scale=True,
                unit_divisor=1024,
        ) as bar:
            for data in response.iter_content(1024):
                file.write(data)
                bar.update(len(data))

    def __apply_predictions(
            G: BaseESGraph,
            predictions: Dict[
                Union[RootEntity, str],
                List[Union[Triple, Tuple[str, str, str]]]
            ]
    ):
        if predictions is None:
            return
        elif isinstance(predictions, dict):
            G.clear_summaries()
            for root_entity, triples in predictions.items():
                G.mark_triples_as_summaries(root_entity, triples)
        else:
            raise ValueError("Predictions should be a dictionary of root entities and their predicted triples.")

    def load_graph(
            self,
            implementation_class: Type[T],
            dataset: DatasetName,
            root_entity_formatter: Optional[callable] = None,
            entity_formatter: Optional[Callable] = None,
            predicate_formatter: Optional[Callable] = None,
            triple_formatter: Optional[Callable] = None) -> T:
        if not issubclass(implementation_class, BaseESGraph):
            raise ValueError("Please use a valid WikESGraph class.")

        if isinstance(dataset, self.WikES_datasets) and not issubclass(implementation_class, WikESBaseGraph):
            raise ValueError("To use a WikES dataset, use WikESGraph or PandasWikESGraph.")
        if isinstance(dataset, self.ESBM_datasets) and not issubclass(implementation_class, ESBMBaseGraph):
            raise ValueError("To use an ESBM dataset you need to use the ESBMGraph class.")

        if issubclass(implementation_class, WikESGraph):
            if not isinstance(dataset, DatasetName):
                raise ValueError("Please use a valid dataset version class.")
            if root_entity_formatter is not None and not isinstance(root_entity_formatter, types.FunctionType):
                raise ValueError("Please provide a valid root entity formatter functon.")
            if entity_formatter is not None and not isinstance(entity_formatter, types.FunctionType):
                raise ValueError("Please provide a valid entity formatter functon.")
            if predicate_formatter is not None and not isinstance(predicate_formatter, types.FunctionType):
                raise ValueError("Please provide a valid predicate formatter functon.")
            if triple_formatter is not None and not isinstance(triple_formatter, types.FunctionType):
                raise ValueError("Please provide a valid triple formatter functon.")
        elif issubclass(implementation_class, PandasWikESGraph) and (
                entity_formatter or predicate_formatter or triple_formatter):
            logger.warning("PandasWikESGraph does not support custom formatters. Ignoring formater functions.")

        dataset_path = self.save_path / dataset.get_version() / f"{dataset.value}.pkl"
        if not dataset_path.exists():
            self.__download_graph(dataset)
        if not dataset_path.exists():
            raise FileNotFoundError(f"Dataset [{dataset}] could not be downloaded.")

        with open(dataset_path, 'rb') as f:
            G: nx.MultiDiGraph = pickle.load(f)
        if not G:
            raise ValueError("Could not load the graph from the dataset file.")
        else:
            logger.debug(f"Graph [{dataset}] file loaded successfully.")

        if issubclass(implementation_class, WikESGraph):
            return WikESGraph(
                G,
                dataset,
                root_entity_formatter, entity_formatter, predicate_formatter, triple_formatter
            )
        elif issubclass(implementation_class, PandasWikESGraph):
            return PandasWikESGraph(G, dataset)
        elif issubclass(implementation_class, ESBMGraph):
            return ESBMGraph(G, dataset, root_entity_formatter, entity_formatter, predicate_formatter, triple_formatter)
        elif issubclass(implementation_class, PandasESBMGraph):
            return PandasESBMGraph(G, dataset)
        else:
            raise ValueError("Please provide a valid Graph class.")

    def load_all_graphs(self,
                        implementation_class: Type[T],
                        dataset_version: Type[DatasetVersion],
                        entity_formatter: Optional[callable] = None,
                        predicate_formatter: Optional[callable] = None,
                        triple_formatter: Optional[callable] = None
                        ) -> Tuple[DatasetName, T]:
        if not issubclass(dataset_version, DatasetVersion):
            raise ValueError("Please use one of the provided version classes.")

        for name, obj in inspect.getmembers(dataset_version):
            if inspect.isclass(obj) and issubclass(obj, DatasetName):
                for enum_member in obj:
                    yield enum_member, self.load_graph(
                        implementation_class, enum_member, entity_formatter,
                        predicate_formatter, triple_formatter
                    )

    @staticmethod
    def F1(
            G: BaseESGraph,
            k: int = None,
            additional_predictions: Dict[Union[RootEntity, str], List[Union[Triple, Tuple[str, str, str]]]] = None
    ) -> float:
        WikESToolkit.__apply_predictions(G, additional_predictions)

        if G.predications() == 0:
            logger.error("No predictions found or provided.")
            return 0

        root_entity_ids = list(G.root_entity_ids())
        n = len(root_entity_ids)
        if n == 0:
            logger.error("No root entities found.")
            return 0

        logger.debug(f"Calculating F1 score...")

        if isinstance(G, ESBMBaseGraph):
            return WikESToolkit._ESBM_F1(G, k)
        elif isinstance(G, WikESBaseGraph):
            return WikESToolkit._WikES_MAP(G, k)
        else:
            raise NotImplementedError("F1 score is not implemented for this graph type.")

    @staticmethod
    def MAP(
            G: BaseESGraph,
            k: int = None,
            additional_predictions: Dict[Union[RootEntity, str], List[Union[Triple, Tuple[str, str, str]]]] = None
    ) -> float:
        WikESToolkit.__apply_predictions(G, additional_predictions)

        if len(G.predications()) == 0:
            logger.error("No predictions found or provided.")
            return 0

        root_entity_ids = list(G.root_entity_ids())
        n = len(root_entity_ids)
        if n == 0:
            logger.error("No root entities found.")
            return 0

        logger.debug(f"Calculating MAP score...")

        if isinstance(G, ESBMBaseGraph):
            return WikESToolkit._ESBM_MAP(G, k)
        elif isinstance(G, WikESBaseGraph):
            return WikESToolkit._WikES_MAP(G, k)
        else:
            raise NotImplementedError("MAP score is not implemented for this graph type.")

    @staticmethod
    def _WikES_F1(G: WikESBaseGraph, k: int) -> float:
        evaluator = WikESSummaryEvaluator(G.root_entity_ids(), G.all_ground_truth_triple_ids(k), G.predications(), k)
        return evaluator.evaluate_f1()

    @staticmethod
    def _WikES_MAP(G: WikESBaseGraph, k: int) -> float:
        evaluator = WikESSummaryEvaluator(G.root_entity_ids(), G.all_ground_truth_triple_ids(k), G.predications(), k)
        return evaluator.evaluate_map()

    @staticmethod
    def _ESBM_F1(G: ESBMBaseGraph, k: int) -> float:
        if k != 5 and k != 10:
            raise ValueError("Only top-5 and top-10 summaries are allowed for ESBM.")

        evaluator = ESBMSummaryEvaluator(G.root_entity_ids(), G.all_gold_top_k(k), G.predications(), k)
        return evaluator.evaluate_f1()

    @staticmethod
    def _ESBM_MAP(G: ESBMBaseGraph, k: int = 5) -> float:
        if k != 5 and k != 10:
            raise ValueError("Only top-5 and top-10 summaries are allowed for ESBM.")

        evaluator = ESBMSummaryEvaluator(G.root_entity_ids(), G.all_gold_top_k(k), G.predications(), k)
        return evaluator.evaluate_map()



from .toolkit import WikESToolkit

from wikes_toolkit.esbm.esbm_versions import ESBMVersions
from wikes_toolkit.esbm.esbm_graph import ESBMGraph as ESBMGraph
from wikes_toolkit.esbm.esbm_pandas_graph import PandasESBMGraph as PandasESBMGraph
from wikes_toolkit.esbm.esbm_graph_components import ESBMEntity, ESBMPredicate, ESBMRootEntity

from wikes_toolkit.wikes.wikes_versions import WikESVersions
from wikes_toolkit.wikes.wikes_graph import WikESGraph
from wikes_toolkit.wikes.wikes_pandas_graph import PandasWikESGraph
from wikes_toolkit.wikes.wikies_graph_components import WikiEntity, WikiPredicate, WikiRootEntity



import logging
from collections import defaultdict
from typing import Union, Tuple, List, Optional, Dict

import networkx as nx

from wikes_toolkit.base.graph_components import Entity, RootEntity, Triple, Predicate
from wikes_toolkit.base.versions import DatasetName
from wikes_toolkit.wikes.wikies_graph_components import WikESBaseGraph, WikiEntity, WikiRootEntity, WikiPredicate, \
    WikiTriple

logger = logging.getLogger(__name__)


class WikESGraph(WikESBaseGraph):

    def __init__(self,
                 G: nx.MultiDiGraph,
                 dataset: DatasetName,
                 root_entity_formatter: Optional[callable] = None,
                 entity_formatter: Optional[callable] = None,
                 predicate_formatter: Optional[callable] = None,
                 triple_formatter: Optional[callable] = None
                 ):
        super().__init__(G, dataset, root_entity_formatter, entity_formatter, predicate_formatter, triple_formatter)

    def _initialize(self):
        self._entities: Dict[str, WikiEntity] = {}
        self._root_entities: Dict[str, WikiRootEntity] = {}
        self._predicates: Dict[str, WikiPredicate] = {}
        self._triples: Dict[Tuple[str, str, str], WikiTriple] = {}
        self._ground_truths: Dict[str, List[Tuple[str, str, str]]] = defaultdict(list)
        logger.debug("Initializing WikESGraph...")
        for node, data in self._G.nodes(data=True):
            if data.get('is_root', False):
                root_entity = WikiRootEntity(
                    identifier=node,
                    wikidata_label=data.get('wikidata_label'),
                    wikidata_description=data.get('wikidata_desc'),
                    wikipedia_id=data.get('wikipedia_id'),
                    wikipedia_title=data.get('wikipedia_title'),
                    category=data.get('category'),
                    str_formatter=self._root_entity_formatter
                )
                self._root_entities[node] = root_entity

            entity = WikiEntity(
                identifier=node,
                wikidata_label=data.get('wikidata_label'),
                wikidata_description=data.get('wikidata_desc'),
                wikipedia_id=data.get('wikipedia_id'),
                wikipedia_title=data.get('wikipedia_title'),
                str_formatter=self._entity_formatter
            )
            self._entities[node] = entity
        logger.debug(f"Entities: {len(self._entities)} initialized.")

        logger.debug("Initializing triples...")
        for u, v, data in self._G.edges(data=True):
            predicate_id = data['predicate']
            if predicate_id not in self._predicates:
                self._predicates[predicate_id] = WikiPredicate(
                    predicate_id=predicate_id,
                    label=data.get('predicate_label'),
                    description=data.get('predicate_desc'),
                    str_formatter=self._predicate_formatter
                )

            triple = WikiTriple(
                subject_entity=self._entities[u],
                predicate=self._predicates[predicate_id],
                object_entity=self._entities[v],
                str_formatter=self._triple_formatter
            )
            self._triples[(u, predicate_id, v)] = triple
            if 'summary_for' in data:
                self._ground_truths[data['summary_for']].append(
                    (
                        triple.subject_entity.identifier,
                        triple.predicate.predicate_id,
                        triple.object_entity.identifier
                    )
                )
        logger.debug(f"Triples: {len(self._triples)} initialized.")

    def root_entities(self) -> List[WikiRootEntity]:
        return super().root_entities()

    def entities(self) -> List[WikiEntity]:
        return super().entities()

    def triples(self) -> List[WikiTriple]:
        return super().triples()

    def predicates(self) -> List[WikiPredicate]:
        return super().predicates()

    def fetch_entity(self, entity: Union[Entity, str]) -> WikiEntity:
        return super().fetch_entity(entity)

    def fetch_root_entity(self, entity: Union[RootEntity, str]) -> WikiRootEntity:
        return super().fetch_root_entity(entity)

    def fetch_predicate(self, predicate: Union[Predicate, str]) -> WikiPredicate:
        return super().fetch_predicate(predicate)

    def fetch_triple(self, triple: Union[
        Triple, Tuple[
            Union[Entity, RootEntity, str],
            Union[Predicate, str],
            Union[Entity, RootEntity, str]
        ],

    ]) -> WikiTriple:
        return super().fetch_triple(triple)

    def neighbors(self, entity: Union[Entity, str]) -> List[WikiTriple]:
        return super().neighbors(entity)



from __future__ import annotations

import logging
from abc import abstractmethod, ABC
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Optional, Union, Dict, Tuple, Callable, List, TypeVar, Type

import networkx as nx
import pandas as pd

from wikes_toolkit.base.versions import DatasetName

logger = logging.getLogger(__name__)


@dataclass
class Entity:
    identifier: str
    str_formatter: Optional[Callable[[Entity], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"Entity(identifier={self.identifier})"

    def __eq__(self, other):
        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)


@dataclass
class RootEntity(Entity):
    category: Optional[str] = None
    str_formatter: Optional[Callable[[RootEntity], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"RootEntity(identifier={self.identifier})"

    def __eq__(self, other):
        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)


@dataclass
class Predicate:
    predicate_id: str
    str_formatter: Optional[Callable[[Predicate], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"Predicate(predicate_id={self.predicate_id})"

    def __eq__(self, other):
        return self.predicate_id == other.predicate_id

    def __hash__(self):
        return hash(self.predicate_id)


@dataclass
class Triple:
    subject_entity: Entity
    predicate: Predicate
    object_entity: Entity
    str_formatter: Optional[Callable[[Triple], str]] = field(default=None, repr=False)

    def key(self):
        return self.subject_entity.identifier, self.predicate.predicate_id, self.object_entity.identifier

    def __iter__(self):
        return iter([self.subject_entity, self.predicate, self.object_entity])

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"Triple(({self.subject_entity.identifier})-[{self.predicate.predicate_id}]->({self.object_entity.identifier})"

    def __eq__(self, other):
        return (
                self.subject_entity.identifier == other.subject_entity.identifier
                and self.predicate.predicate_id == other.predicate.predicate_id
                and self.object_entity.identifier == other.object_entity.identifier
        )

    def __hash__(self):
        return hash(self.key())


ROOT_E = TypeVar('ROOT_E', bound=RootEntity)
E = TypeVar('E', bound=Entity)
T = TypeVar('T', bound=Triple)
P = TypeVar('P', bound=Predicate)


class BaseESGraph(ABC):
    _dataset_name: DatasetName
    _entities: Union[Dict[str, Entity], pd.DataFrame]
    _root_entities: Union[Dict[str, RootEntity], pd.DataFrame]
    _predicates: Union[Dict[str, Predicate], pd.DataFrame]
    _triples: Union[Dict[Tuple[str, str, str], Triple], pd.DataFrame]
    _ground_truths: Union[Dict[str, List[Tuple[str, str, str]]], pd.DataFrame]
    _predicted_summaries: Dict[str, List[Tuple[str, str, str]]] = defaultdict(list)
    _G: nx.MultiDiGraph
    _root_entity_formatter: Callable
    _entity_formatter: Callable
    _predicate_formatter: Callable
    _triple_formatter = Callable

    def __init__(self, G: nx.MultiDiGraph, dataset: DatasetName,
                 root_type: Type, entity_type: Type,
                 predicate_type: Type, triple_type: Type,
                 root_entity_formatter: Optional[callable] = None,
                 entity_formatter: Optional[callable] = None,
                 predicate_formatter: Optional[callable] = None,
                 triple_formatter: Optional[callable] = None):
        self._G: nx.MultiDiGraph = G
        self._dataset_name = dataset
        self._root_type = root_type
        self._entity_type = entity_type
        self._predicate_type = predicate_type
        self._triple_type = triple_type
        self._root_entity_formatter = root_entity_formatter
        self._entity_formatter = entity_formatter
        self._predicate_formatter = predicate_formatter
        self._triple_formatter = triple_formatter
        self._initialize()

    @abstractmethod
    def _initialize(self):
        pass

    @abstractmethod
    def root_entities(self):
        if isinstance(self._root_entities, pd.DataFrame):
            return self._root_entities
        else:
            return list(self._root_entities.values())

    def root_entity_ids(self) -> List[str]:
        if isinstance(self._root_entities, pd.DataFrame):
            return self._root_entities.index.tolist()
        else:
            return list(self._root_entities.keys())

    @abstractmethod
    def entities(self):
        if isinstance(self._entities, pd.DataFrame):
            return self._entities
        else:
            return list(self._entities.values())

    @abstractmethod
    def triples(self):
        if isinstance(self._triples, pd.DataFrame):
            return self._triples
        else:
            return list(self._triples.values())

    @abstractmethod
    def predicates(self):
        if isinstance(self._predicates, pd.DataFrame):
            return self._predicates
        else:
            return list(self._predicates.values())

    def total_entities(self) -> int:
        if isinstance(self._entities, pd.DataFrame):
            return self._entities.shape[0]
        else:
            return len(self._entities)

    def total_triples(self) -> int:
        if isinstance(self._triples, pd.DataFrame):
            return self._triples.shape[0]
        else:
            return len(self._triples)

    @abstractmethod
    def fetch_entity(self, entity: Union[Entity, str, pd.Series]):

        if isinstance(self._entities, pd.DataFrame):
            try:
                if isinstance(entity, pd.Series):
                    entity = entity.name
                entity = self._entities.loc[entity]
            except KeyError:
                raise ValueError(f"Entity with identifier: {entity} not found.")
            return entity
        if isinstance(entity, self._root_type) or isinstance(entity, self._entity_type):
            identifier = entity.identifier
        else:
            identifier = entity
        if identifier not in self._entities:
            raise ValueError(f"Entity with identifier: {identifier} not found.")
        return self._entities[identifier]

    @abstractmethod
    def fetch_root_entity(self, entity: Union[RootEntity, str, pd.Series]):
        if isinstance(self._root_entities, pd.DataFrame):
            try:
                if isinstance(entity, pd.Series):
                    entity = entity.name
                entity = self._root_entities.loc[entity]
            except KeyError:
                raise ValueError(f"Entity with identifier: {entity} is not a root entity.")
            return entity
        else:
            if isinstance(entity, self._root_type) or isinstance(entity, self._entity_type):
                identifier = entity.identifier
            else:
                identifier = entity
            if identifier not in self._root_entities:
                raise ValueError(f"Entity with identifier: {identifier} not found in root entities.")
            return self._root_entities[identifier]

    def fetch_root_entity_id(self, entity: Union[RootEntity, str, pd.Series]) -> str:
        return self.fetch_root_entity(entity).identifier

    @abstractmethod
    def fetch_predicate(self, predicate: Union[Predicate, str, pd.Series]):
        if isinstance(self._predicates, pd.DataFrame):
            try:
                if isinstance(predicate, pd.Series):
                    predicate = predicate.name
                predicate = self._predicates.loc[predicate]
            except KeyError:
                raise ValueError(f"Predicate with predicate_id: {predicate} not found")
            return predicate
        else:
            if isinstance(predicate, self._predicate_type):
                predicate_id = predicate.predicate_id
            else:
                predicate_id = predicate

            if predicate_id not in self._predicates:
                raise ValueError(f"Predicate with predicate_id: {predicate_id} not found")
            return self._predicates[predicate_id]

    @abstractmethod
    def fetch_triple(self, triple: Union[
        Triple, Tuple[
            Union[Entity, str],
            Union[Predicate, str],
            Union[Entity, str]
        ],
        pd.Series
    ]):
        if isinstance(self._triples, pd.DataFrame):
            if isinstance(triple, pd.Series):
                triple = triple['subject'], triple['predicate'], triple['object']
            fetched_triple = self._triples[
                (self._triples['subject'] == triple[0]) &
                (self._triples['predicate'] == triple[1]) &
                (self._triples['object'] == triple[2])
                ]
            if fetched_triple.empty:
                raise ValueError(f"Triple ({triple}) not found.")
            return fetched_triple.iloc[0]
        else:
            if isinstance(triple, Tuple):
                subject_entity = self.fetch_entity(triple[0])
                predicate = self.fetch_predicate(triple[1])
                object_entity = self.fetch_entity(triple[2])
                triple_key = (subject_entity.identifier, predicate.predicate_id, object_entity.identifier)
            else:
                triple_key = (
                    triple.subject_entity.identifier,
                    triple.predicate.predicate_id,
                    triple.object_entity.identifier
                )
            if triple_key not in self._triples:
                raise ValueError(f"Triple {triple_key} not found.")
            return self._triples[triple_key]

    def fetch_triple_ids(self, triple: Union[
        Triple, Tuple[
            Union[Entity, str],
            Union[Predicate, str],
            Union[Entity, str]
        ],
        pd.Series
    ]) -> Tuple[str, str, str]:
        if isinstance(self._triples, pd.DataFrame):
            return self.fetch_triple(triple)[['subject', 'predicate', 'object']].apply(tuple, axis=1)
        else:
            triple = self.fetch_triple(triple)
            return triple.subject_entity.identifier, triple.predicate.predicate_id, triple.object_entity.identifier

    @abstractmethod
    def neighbors(self, entity: Union[Entity, str, pd.Series]):
        if isinstance(self._entities, pd.DataFrame):
            entity = self.fetch_entity(entity)
            return self._triples[
                (self._triples['subject'] == entity.name) |
                (self._triples['object'] == entity.name)
                ]
        else:
            entity = self.fetch_entity(entity)
            entity_id = entity.identifier
            neighbors = [
                self._triples[(u, data['predicate'], v)]
                for u, v, data in self._G.edges(entity_id, data=True)
            ]
            neighbors.extend([
                self._triples[(u, data['predicate'], v)]
                for u, v, data in self._G.in_edges(entity_id, data=True)
            ])
            return neighbors

    def degree(self, entity: Union[Entity, str, pd.Series]) -> int:
        if isinstance(self._entities, pd.DataFrame):
            return self.neighbors(entity).shape[0]
        else:
            entity = self.fetch_entity(entity)
            return self._G.degree(entity.identifier)

    def _add_predication_if_not_exists(self, root_entity: str, triple: Tuple[str, str, str]):
        if triple not in self._predicted_summaries[root_entity]:
            self._predicted_summaries[root_entity].append(triple)

    def mark_triple_as_summary(self, root_entity: Union[RootEntity, str, pd.Series], triple: Union[
        str,
        Triple,
        Tuple[
            Union[Entity, str],
            Union[Predicate, str],
            Union[Entity, str]
        ],
        pd.Series
    ]):
        if isinstance(self._entities, pd.DataFrame):
            identifier = str(self.fetch_root_entity(root_entity).name)
            triple = self.fetch_triple(triple)

            if identifier != triple['subject'] and identifier != triple['object']:
                raise ValueError(
                    f"Root entity: {identifier} should be either a subject or an object of the triple in a summary.")

            self._add_predication_if_not_exists(identifier, (triple['subject'], triple['predicate'], triple['object']))
        else:
            root_entity_id = self.fetch_entity(root_entity).identifier
            triple = self.fetch_triple(triple)
            if root_entity_id not in {triple.subject_entity.identifier, triple.object_entity.identifier}:
                raise ValueError(
                    f"Root entity: {root_entity_id} should be either a subject or an object of the triple in a summary.")

            self._add_predication_if_not_exists(
                root_entity_id,
                (
                    triple.subject_entity.identifier,
                    triple.predicate.predicate_id,
                    triple.object_entity.identifier
                )
            )

    def clear_summaries(self):
        self._predicted_summaries.clear()

    def mark_triples_as_summaries(
            self,
            root_entity: Union[RootEntity, str, pd.Series],
            triples: List[Union[str, Triple, Union[Triple, Tuple[str, str, str], pd.DataFrame]]]
    ):
        if isinstance(triples, (Tuple, pd.Series)):
            return self.mark_triple_as_summary(root_entity, triples)
        if isinstance(triples, pd.DataFrame):
            for index, triple in triples.iterrows():
                self.mark_triple_as_summary(root_entity, triple)
        else:
            for triple in triples:
                self.mark_triple_as_summary(root_entity, triple)

    def predications(self) -> Dict[str, List[Tuple[str, str, str]]]:
        return self._predicted_summaries

    def predications_for_root(self, identifier: str) -> List[Tuple[str, str, str]]:
        if identifier in self._predicted_summaries:
            return self._predicted_summaries[identifier]
        else:
            return list()

    @abstractmethod
    def f1_score(self, top_k: int = None, no_rel: bool = False):
        pass

    @abstractmethod
    def map_score(self, top_k: int = None, no_rel: bool = False):
        pass



from __future__ import annotations

import logging
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Optional, Union, Tuple, Set, Callable, List, Dict

import networkx as nx
import pandas as pd

from wikes_toolkit.base.graph_components import Entity, Predicate, Triple, BaseESGraph, RootEntity
from wikes_toolkit.base.versions import DatasetName
from wikes_toolkit.esbm.esbm_eval import ESBMSummaryEvaluator

logger = logging.getLogger(__name__)


@dataclass
class ESBMEntity(Entity):
    identifier: str
    str_formatter: Optional[Callable[[ESBMEntity], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"ESBMEntity(identifier={self.identifier})"

    def __eq__(self, other):
        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)


@dataclass
class ESBMRootEntity(ESBMEntity):
    eid: int = None
    category: Optional[str] = None
    str_formatter: Optional[Callable[[ESBMRootEntity], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"ESBMRootEntity(identifier={self.identifier}, eid:{self.eid}, category={self.category})"

    def __eq__(self, other):
        if isinstance(other, str):
            return self.identifier == other
        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)


@dataclass
class ESBMPredicate(Predicate):
    predicate_id: str
    str_formatter: Optional[Callable[[ESBMPredicate], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"ESBMPredicate(predicate_id={self.predicate_id})"

    def __eq__(self, other):
        return self.predicate_id == other.predicate_id

    def __hash__(self):
        return hash(self.predicate_id)


@dataclass
class ESBMTriple(Triple):
    subject_entity: ESBMEntity
    predicate: ESBMPredicate
    object_entity: ESBMEntity
    str_formatter: Optional[Callable[[Triple], str]] = field(default=None, repr=False)

    def key(self):
        return self.subject_entity.identifier, self.predicate.predicate_id, self.object_entity.identifier

    def __iter__(self):
        return iter([self.subject_entity, self.predicate, self.object_entity])

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"ESBMTriple(({self.subject_entity.identifier})-[{self.predicate.predicate_id}]->({self.object_entity.identifier})"

    def __eq__(self, other):
        return (
                self.subject_entity.identifier == other.subject_entity.identifier
                and self.predicate.predicate_id == other.predicate.predicate_id
                and self.object_entity.identifier == other.object_entity.identifier
        )

    def __hash__(self):
        return hash(self.key())


class ESBMBaseGraph(BaseESGraph):
    _dataset_name: DatasetName

    def __init__(self, G: nx.MultiDiGraph, dataset: DatasetName,
                 root_entity_formatter: Optional[callable] = None,
                 entity_formatter: Optional[callable] = None,
                 predicate_formatter: Optional[callable] = None,
                 triple_formatter: Optional[callable] = None):
        super().__init__(
            G, dataset,
            ESBMRootEntity, ESBMEntity, ESBMTriple, ESBMPredicate,
            root_entity_formatter, entity_formatter, predicate_formatter, triple_formatter
        )

    @abstractmethod
    def all_gold_top_k(self, k: int) -> Dict[str, List[List[Tuple[str, str, str]]]]:
        pass

    @abstractmethod
    def gold_top_5(self, root_entity: Union[ESBMRootEntity, str, pd.Series], annotator_index: int) -> Union[
        List[ESBMTriple], pd.DataFrame
    ]:
        pass

    @abstractmethod
    def gold_top_10(self, root_entity: Union[ESBMRootEntity, str, pd.Series], annotator_index: int) -> Union[
        List[ESBMTriple], pd.DataFrame
    ]:
        pass

    def f1_score(self, k: int = None, no_rel: bool = False):
        if k is None:
            raise ValueError("top_k should be provided for ESBM")
        if k not in [5, 10]:
            raise ValueError("k should be 5 or 10")
        return ESBMSummaryEvaluator(
            super().root_entities(),
            self.all_gold_top_k(k),
            self._predicted_summaries,
            k
        ).evaluate_f1(no_rel)

    def map_score(self, k: int = None, no_rel: bool = False):
        if k is None:
            raise ValueError("top_k should be provided for ESBM")
        if k not in [5, 10]:
            raise ValueError("k should be 5 or 10")
        return ESBMSummaryEvaluator(
            super().root_entities(),
            self.all_gold_top_k(k),
            self._predicted_summaries,
            k
        ).evaluate_map(no_rel)



from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum


@dataclass
class DatasetVersion:
    base_url: str


class DatasetName(Enum):

    @abstractmethod
    def get_dataset_url(self) -> str:
        pass

    @abstractmethod
    def get_version(self) -> str:
        pass



from wikes_toolkit.base.versions import DatasetVersion, DatasetName


class ESBMVersions(DatasetVersion):
    base_url = 'https://github.com/msorkhpar/ESBM-to-nx-format/releases/download/ESBM/'

    @staticmethod
    def available_versions():
        subclasses = set()

        def get_all_subclasses(cls):
            for attr_name in dir(cls):
                attr = getattr(cls, attr_name)
                if isinstance(attr, type) and issubclass(attr, DatasetVersion):
                    get_all_subclasses(attr)
                elif isinstance(attr, type) and issubclass(attr, DatasetName):
                    subclasses.add(attr)

        get_all_subclasses(ESBMVersions)
        return tuple(subclasses)

    @staticmethod
    def get_dataset_url(dataset: DatasetName) -> str:
        return ESBMVersions.base_url + dataset.value + ".pkl"

    class V1Dot0(DatasetName):
        DBPEDIA_FULL = 'v1_0_dbpedia_full'
        LMDB_FULL = 'v1_0_lmdb_full'

        def get_dataset_url(self):
            return ESBMVersions.get_dataset_url(self)

        def get_version(self) -> str:
            return 'esbm-1.0'

    class V1Dot1(DatasetName):
        DBPEDIA_FULL = 'v1_1_dbpedia_full'
        LMDB_FULL = 'v1_1_lmdb_full'

        def get_dataset_url(self):
            return ESBMVersions.get_dataset_url(self)

        def get_version(self) -> str:
            return 'esbm-1.1'

    class V1Dot2(DatasetName):
        DBPEDIA_FULL = 'v1_2_dbpedia_full'
        LMDB_FULL = 'v1_2_linkedmdb_full'

        DBPEDIA_TRAIN_0 = 'v1_2_dbpedia_train_0'
        DBPEDIA_VALIDATION_0 = 'v1_2_dbpedia_val_0'
        DBPEDIA_TEST_0 = 'v1_2_dbpedia_test_0'
        LMDB_TRAIN_0 = 'v1_2_linkedmdb_train_0'
        LMDB_VALIDATION_0 = 'v1_2_linkedmdb_val_0'
        LMDB_TEST_0 = 'v1_2_linkedmdb_test_0'

        DBPEDIA_TRAIN_1 = 'v1_2_dbpedia_train_1'
        DBPEDIA_VALIDATION_1 = 'v1_2_dbpedia_val_1'
        DBPEDIA_TEST_1 = 'v1_2_dbpedia_test_1'
        LMDB_TRAIN_1 = 'v1_2_lmdb_full_1'
        LMDB_VALIDATION_1 = 'v1_2_linkedmdb_val_1'
        LMDB_TEST_1 = 'v1_2_linkedmdb_test_1'

        DBPEDIA_TRAIN_2 = 'v1_2_dbpedia_train_2'
        DBPEDIA_VALIDATION_2 = 'v1_2_dbpedia_val_2'
        DBPEDIA_TEST_2 = 'v1_2_dbpedia_test_2'
        LMDB_TRAIN_2 = 'v1_2_linkedmdb_train_2'
        LMDB_VALIDATION_2 = 'v1_2_linkedmdb_val_2'
        LMDB_TEST_2 = 'v1_2_linkedmdb_test_2'

        DBPEDIA_TRAIN_3 = 'v1_2_dbpedia_train_3'
        DBPEDIA_VALIDATION_3 = 'v1_2_dbpedia_val_3'
        DBPEDIA_TEST_3 = 'v1_2_dbpedia_test_3'
        LMDB_TRAIN_3 = 'v1_2_linkedmdb_train_3'
        LMDB_VALIDATION_3 = 'v1_2_linkedmdb_val_3'
        LMDB_TEST_3 = 'v1_2_linkedmdb_test_3'

        DBPEDIA_TRAIN_4 = 'v1_2_dbpedia_train_4'
        DBPEDIA_VALIDATION_4 = 'v1_2_dbpedia_val_4'
        DBPEDIA_TEST_4 = 'v1_2_dbpedia_test_4'
        LMDB_TRAIN_4 = 'v1_2_linkedmdb_train_4'
        LMDB_VALIDATION_4 = 'v1_2_linkedmdb_val_4'
        LMDB_TEST_4 = 'v1_2_linkedmdb_test_4'

        def get_dataset_url(self):
            return ESBMVersions.get_dataset_url(self)

        def get_version(self) -> str:
            return 'esbm-1.2'

    class Plus(DatasetName):
        DBPEDIA_FULL = 'Plus_dbpedia_full'
        LMDB_FULL = 'Plus_linkedmdb_full'

        def get_dataset_url(self):
            return ESBMVersions.get_dataset_url(self)

        def get_version(self) -> str:
            return 'esbm-plus'



import logging
import os
import shutil

import networkx as nx
import pandas as pd
from filehash import FileHash

from wikes_toolkit.base.versions import DatasetName

logger = logging.getLogger(__name__)


def assign_ids(df, new_identifier_col_name=None):
    df['id'] = range(len(df))
    df.reset_index(inplace=True)
    df.set_index('id', inplace=True)
    if new_identifier_col_name:
        df.rename(columns={'identifier': new_identifier_col_name}, inplace=True)


def map_column_to_ids(df, column, mapping):
    df[column] = df[column].map(mapping)


def export(
        output_path: str,
        license_path: str,
        G: nx.MultiDiGraph,
        dataset: DatasetName,
        entities: pd.DataFrame,
        root_nodes: pd.DataFrame,
        triples: pd.DataFrame, predicates: pd.DataFrame, ground_truths: pd.DataFrame) -> str:
    dataset_name = dataset.value
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    assign_ids(entities, 'entity')
    assign_ids(root_nodes, 'entity')
    assign_ids(predicates, 'predicate')
    assign_ids(ground_truths, 'root_entity')

    map_entity_to_id = {v: k for k, v in entities['entity'].to_dict().items()}
    map_predicate_to_id = {v: k for k, v in predicates['predicate'].to_dict().items()}

    map_column_to_ids(root_nodes, 'entity', map_entity_to_id)
    map_column_to_ids(triples, 'subject', map_entity_to_id)
    map_column_to_ids(triples, 'object', map_entity_to_id)
    map_column_to_ids(triples, 'predicate', map_predicate_to_id)
    map_column_to_ids(ground_truths, 'root_entity', map_entity_to_id)
    map_column_to_ids(ground_truths, 'subject', map_entity_to_id)
    map_column_to_ids(ground_truths, 'object', map_entity_to_id)
    map_column_to_ids(ground_truths, 'predicate', map_predicate_to_id)

    root_nodes.drop(
        columns=['wikidata_label', 'wikidata_description', 'wikipedia_id', 'wikipedia_title'],
        inplace=True
    )

    output_csv_path = os.path.join(output_path, dataset_name)
    os.makedirs(output_csv_path, exist_ok=True)

    logger.info(f"Exporting to {output_csv_path}")

    entities.to_csv(os.path.join(output_csv_path, f'{dataset_name}-entities.csv'))
    predicates.to_csv(os.path.join(output_csv_path, f'{dataset_name}-predicates.csv'))
    root_nodes.to_csv(os.path.join(output_csv_path, f'{dataset_name}-root-entities.csv'), index=False)
    triples.to_csv(os.path.join(output_csv_path, f'{dataset_name}-triples.csv'), index=False)
    ground_truths.to_csv(os.path.join(output_csv_path, f'{dataset_name}-ground-truths.csv'),
                         index=False)
    shutil.copy(license_path, os.path.join(output_csv_path, 'LICENSE'))
    shutil.make_archive(output_csv_path, 'zip', output_csv_path)
    shutil.rmtree(output_csv_path)
    nx.write_graphml(G, os.path.join(output_path, f'{dataset_name}.graphml'))

    logger.debug(f"Exported {dataset_name} to {output_csv_path} successfully.")
    return FileHash('sha256').hash_file(os.path.join(output_path, f'{dataset_name}.zip'))



from typing import Dict, List, Tuple
from wikes_toolkit.base.evaluate import f1, map, f1_norel, map_norel


class ESBMSummaryEvaluator:
    def __init__(self, root_entities: List[str], gold_summaries: Dict[str, List[List[Tuple[str, str, str]]]],
                 predictions: Dict[str, List[Tuple[str, str, str]]], top_k: int):
        self.gold_summaries = gold_summaries
        self.predictions = predictions
        self.root_entities = root_entities
        self.top_k = top_k

    def get_gold_summaries(self, entity_id: str) -> List[List[Tuple[str, str, str]]]:
        result = list()
        summaries = self.gold_summaries.get(entity_id, [])
        for summary in summaries:
            result.append(summary[:self.top_k])
        return result

    def get_predications(self, entity_id: str) -> List[Tuple[str, str, str]]:
        predictions = self.predictions.get(entity_id, [])
        return predictions[:self.top_k]

    def evaluate_f1(self, no_rel: bool = False):
        score_function = f1_norel if no_rel else f1

        def calculate_average_prf(gold_summaries: List[List[Tuple[str, str, str]]],
                                  algo_summary: List[Tuple[str, str, str]]):
            f1_sum = 0
            user_count = len(gold_summaries)

            for gold_summary in gold_summaries:
                f1_sum += score_function(gold_summary, algo_summary)

            return f1_sum / user_count

        entity_count = len(self.root_entities)
        dataset_f1_sum = 0

        for entity_id in self.root_entities:
            gold_summaries = self.get_gold_summaries(entity_id)
            algo_summary = self.get_predications(entity_id)

            if algo_summary:
                dataset_f1_sum += calculate_average_prf(gold_summaries, algo_summary)

        return dataset_f1_sum / entity_count

    def evaluate_map(self, no_rel: bool = False):
        score_function = map_norel if no_rel else map

        def calculate_average_map(gold_summaries: List[List[Tuple[str, str, str]]],
                                  algo_rank: List[Tuple[str, str, str]]):
            map_sum = 0
            user_count = len(gold_summaries)

            for gold_summary in gold_summaries:
                map_score = score_function(gold_summary, algo_rank)
                map_sum += map_score

            return map_sum / user_count

        entity_count = len(self.root_entities)
        dataset_map_sum = 0

        for entity_id in self.root_entities:
            gold_summaries = self.get_gold_summaries(entity_id)
            algo_summary = self.get_predications(entity_id)

            if algo_summary:
                dataset_map_sum += calculate_average_map(gold_summaries, algo_summary)

        avg_map = dataset_map_sum / entity_count

        return avg_map



import logging
from collections import defaultdict
from typing import Union, Tuple, Dict, List

import networkx as nx
import pandas as pd

from wikes_toolkit.base.versions import DatasetName
from wikes_toolkit.esbm.esbm_graph_components import ESBMBaseGraph, ESBMRootEntity, ESBMTriple

logger = logging.getLogger(__name__)


class PandasESBMGraph(ESBMBaseGraph):
    def __init__(self, G: nx.MultiDiGraph, dataset: DatasetName):
        super().__init__(G, dataset)

    @staticmethod
    def _build_pandas_gold_top_k(gold_data: Dict[Tuple[str, int], List[Tuple[str, str, str, str]]]):
        df = (pd.DataFrame(
            [
                {
                    'root_entity': key[0],
                    'annotator_index': key[1],
                    'order': inner_item[0],
                    'subject': inner_item[1],
                    'predicate': inner_item[2],
                    'object': inner_item[3]
                } for key, value in gold_data.items() for inner_item in value
            ], columns=[
                'root_entity', 'annotator_index', 'order', 'subject', 'predicate', 'object'
            ])
              .set_index(['root_entity', 'annotator_index', 'order'])
              .sort_index()
              .reset_index(level='order'))
        return df

    def _initialize(self):
        logger.debug("Initializing PandasESBMGraph...")

        nodes_data = [
            {
                'identifier': node,
                'eid': data.get('eid'),
                'label': data.get('label'),
                'category': data.get('category')
            }
            for node, data in self._G.nodes(data=True) if data.get('is_root', False)
        ]
        self._root_entities = pd.DataFrame(nodes_data, columns=[
            'identifier', 'eid', 'label', 'category'
        ]).set_index('identifier')
        del nodes_data
        self._root_entities.sort_values('eid', inplace=True)
        logger.debug(f"Root Entities: {self._root_entities.shape[0]} initialized.")

        nodes_data = [
            {
                'identifier': node,
            }
            for node, data in self._G.nodes(data=True)
        ]
        self._entities = pd.DataFrame(nodes_data, columns=['identifier']).set_index('identifier')
        logger.debug(f"Entities: {self._entities.shape[0]} initialized.")
        del nodes_data

        logger.debug("Initializing triples...")
        edges_data = []
        predicates_data = set()

        gold5_data: Dict[Tuple[str, int], List[Tuple[str, str, str, str]]] = defaultdict(list)
        gold10_data: Dict[Tuple[str, int], List[Tuple[str, str, str, str]]] = defaultdict(list)

        for u, v, data in self._G.edges(data=True):
            predicate_id = data['predicate']
            edges_data.append({'subject': u, 'predicate': predicate_id, 'object': v})

            if predicate_id not in predicates_data:
                predicates_data.add(predicate_id)

            if 'summary_for' in data:
                for i in range(6):
                    if f"in_gold_top5_{i}" in data:
                        order = data[f"in_gold_top5_{i}"]
                        gold5_data[(data['summary_for'], i)].append((order, u, predicate_id, v))
                    if f"in_gold_top10_{i}" in data:
                        order = data[f"in_gold_top10_{i}"]
                        gold10_data[(data['summary_for'], i)].append((order, u, predicate_id, v))

        self._predicates = pd.DataFrame(
            list(predicates_data),
            columns=['identifier']
        ).set_index('identifier')
        del predicates_data

        self._triples = pd.DataFrame(edges_data, columns=['subject', 'predicate', 'object'])
        del edges_data
        logger.debug(f"Triples: {self._triples.shape[0]} initialized.")

        self._gold_top_5 = PandasESBMGraph._build_pandas_gold_top_k(gold5_data)
        del gold5_data

        self._gold_top_10 = PandasESBMGraph._build_pandas_gold_top_k(gold10_data)
        del gold10_data

    def root_entities(self) -> pd.DataFrame:
        return super().root_entities()

    def entities(self) -> pd.DataFrame:
        return super().entities()

    def triples(self) -> pd.DataFrame:
        return super().triples()

    def predicates(self) -> pd.DataFrame:
        return super().predicates()

    def fetch_entity(self, entity: [str, pd.Series]) -> pd.Series:
        return super().fetch_entity(entity)

    def fetch_root_entity(self, entity: [str, pd.Series]) -> pd.Series:
        return super().fetch_root_entity(entity)

    def fetch_predicate(self, predicate: [str, pd.Series]) -> pd.Series:
        return super().fetch_predicate(predicate)

    def fetch_triple(self, triple: Union[Tuple[str, str, str], pd.Series]) -> pd.Series:
        return super().fetch_triple(triple)

    def neighbors(self, entity: [str, pd.Series]) -> pd.DataFrame:
        return super().neighbors(entity)

    def all_gold_top_k(self, k: int) -> Dict[str, List[List[Tuple[str, str, str]]]]:
        if k not in [5, 10]:
            raise ValueError("k should be 5 or 10")
        result: Dict[str, List[List[Tuple[str, str, str]]]] = dict()
        gold_top_k = self._gold_top_5 if k == 5 else self._gold_top_10
        all_golds = gold_top_k.groupby(['root_entity', 'annotator_index']).apply(
            lambda x: tuple(x[['subject', 'predicate', 'object']].values.tolist())
        ).groupby('root_entity').agg(list).to_dict()
        for index, row in self._root_entities.iterrows():
            result[str(index)] = all_golds[index]
        return result

    def gold_top_5(self, root_entity: Union[ESBMRootEntity, str], annotator_index: int) -> pd.DataFrame:
        if annotator_index < 0 or annotator_index > 5:
            raise ValueError("Annotator index should be between 0 and 5.")
        root_entity = self.fetch_root_entity(root_entity).name
        return self._gold_top_5.loc[(root_entity, annotator_index)]

    def gold_top_10(self, root_entity: Union[ESBMRootEntity, str], annotator_index: int) -> pd.DataFrame:
        if annotator_index < 0 or annotator_index > 5:
            raise ValueError("Annotator index should be between 0 and 5.")
        root_entity = self.fetch_root_entity(root_entity).name
        return self._gold_top_10.loc[(root_entity, annotator_index)]

    def export_as_csv(self, path: str):
        raise NotImplementedError("Exporting to CSV is not yet implemented for PandasESBMGraph.")



from typing import Dict, List, Tuple
from wikes_toolkit.base.evaluate import f1, map, f1_norel, map_norel


class WikESSummaryEvaluator:
    def __init__(self, root_entities: List[str], ground_truth: Dict[str, List[Tuple[str, str, str]]],
                 predictions: Dict[str, List[Tuple[str, str, str]]]):
        self.root_entities = root_entities
        self.ground_truth = ground_truth
        self.predictions = predictions

    def get_summaries(self, entity_id: str) -> List[Tuple[str, str, str]]:
        return self.ground_truth.get(entity_id, [])

    def get_predications(self, entity_id: str, top_k: int = None) -> List[Tuple[str, str, str]]:
        predictions = self.predictions.get(entity_id, [])
        ground_truth = self.ground_truth.get(entity_id, [])
        if top_k and top_k < len(ground_truth):
            predictions = predictions[:top_k]
        else:
            predictions = predictions[:len(ground_truth)]
        return predictions

    def evaluate_f1(self, top_k: int = None, no_rel: bool = False):
        score_function = f1_norel if no_rel else f1
        entity_count = len(self.root_entities)
        dataset_f1_sum = 0

        for entity_id in self.root_entities:
            gold_summaries = self.get_summaries(entity_id)
            algo_summary = self.get_predications(entity_id, top_k)

            if algo_summary:
                dataset_f1_sum += score_function(gold_summaries, algo_summary)

        return dataset_f1_sum / entity_count

    def evaluate_map(self, top_k: int = None, no_rel: bool = False):
        score_function = map_norel if no_rel else map
        entity_count = len(self.root_entities)
        dataset_map_sum = 0

        for entity_id in self.root_entities:
            gold_summaries = self.get_summaries(entity_id)
            algo_summary = self.get_predications(entity_id, top_k)

            if algo_summary:
                dataset_map_sum += score_function(gold_summaries, algo_summary)

        avg_map = dataset_map_sum / entity_count

        return avg_map



from __future__ import annotations

import logging
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Optional, Union, Tuple, Callable, List, Dict

import networkx as nx
import pandas as pd

from wikes_toolkit.base.graph_components import Entity, Predicate, Triple, BaseESGraph
from wikes_toolkit.base.versions import DatasetName
from wikes_toolkit.wikes.wikes_eval import WikESSummaryEvaluator

logger = logging.getLogger(__name__)


@dataclass
class WikiEntity(Entity):
    identifier: str
    wikidata_label: Optional[str] = None
    wikidata_description: Optional[str] = None
    wikipedia_id: Optional[int] = None
    wikipedia_title: Optional[str] = None
    str_formatter: Optional[Callable[[WikiEntity], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"Entity(identifier={self.identifier}, wikidata_label={self.wikidata_label})"

    def __eq__(self, other):
        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)


@dataclass
class WikiRootEntity(WikiEntity):
    category: Optional[str] = None
    str_formatter: Optional[Callable[[WikiRootEntity], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"RootEntity(identifier={self.identifier}, wikidata_label={self.wikidata_label})"

    def __eq__(self, other):
        if isinstance(other, str):
            return self.identifier == other
        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)


@dataclass
class WikiPredicate(Predicate):
    predicate_id: str
    label: Optional[str] = None
    description: Optional[str] = None
    str_formatter: Optional[Callable[[WikiPredicate], str]] = field(default=None, repr=False)

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"Predicate(predicate_id={self.predicate_id}, label={self.label})"

    def __eq__(self, other):
        return self.predicate_id == other.predicate_id

    def __hash__(self):
        return hash(self.predicate_id)


@dataclass
class WikiTriple(Triple):
    subject_entity: WikiEntity
    predicate: WikiPredicate
    object_entity: WikiEntity
    str_formatter: Optional[Callable[[Triple], str]] = field(default=None, repr=False)

    def key(self):
        return self.subject_entity.identifier, self.predicate.predicate_id, self.object_entity.identifier

    def __iter__(self):
        return iter([self.subject_entity, self.predicate, self.object_entity])

    def __str__(self):
        if self.str_formatter:
            return self.str_formatter(self)
        return f"Triple(({self.subject_entity.identifier})-[{self.predicate.predicate_id}]->({self.object_entity.identifier})"

    def __eq__(self, other):
        return (
                self.subject_entity.identifier == other.subject_entity.id
                and self.predicate.predicate_id == other.predicate.predicate_id
                and self.object_entity.identifier == other.object_entity.id
        )

    def __hash__(self):
        return hash(self.key())


class WikESBaseGraph(BaseESGraph):
    def __init__(self, G: nx.MultiDiGraph,
                 dataset: DatasetName,
                 root_entity_formatter: Optional[callable] = None,
                 entity_formatter: Optional[callable] = None,
                 predicate_formatter: Optional[callable] = None,
                 triple_formatter: Optional[callable] = None):
        super().__init__(
            G, dataset,
            WikiRootEntity, WikiEntity, WikiTriple, WikiPredicate,
            root_entity_formatter, entity_formatter, predicate_formatter, triple_formatter
        )

    @abstractmethod
    def _initialize(self):
        pass

    def ground_truths(self, root_entity: Union[WikiRootEntity, str, pd.Series]) -> Union[
        List[Tuple[str, str, str]],
        pd.Series
    ]:
        if isinstance(self._ground_truths, pd.DataFrame):
            root_entity = self.fetch_root_entity(root_entity)
            return self._ground_truths[self._ground_truths.index == root_entity.name]
        else:
            root_entity_id = self.fetch_root_entity(root_entity).identifier
            if root_entity_id not in self._ground_truths:
                raise ValueError(f"No ground truth summaries found for root_entity: {root_entity_id}.")
            return self._ground_truths[root_entity_id]

    def ground_truth_triple_ids(self, root_entity: Union[WikiRootEntity, str, pd.Series]) -> List[Tuple[str, str, str]]:
        if isinstance(self._ground_truths, pd.DataFrame):
            return list(self.ground_truths(root_entity)[['subject', 'predicate', 'object']].apply(tuple, axis=1))
        else:
            return [(s, p, o) for s, p, o in self.ground_truths(root_entity)]

    def all_ground_truth_triple_ids(self, k: int = None) -> Dict[str, List[Tuple[str, str, str]]]:
        if k is not None:
            result = {}
            for root_entity in self.root_entity_ids():
                ground_truth_triples = self.ground_truth_triple_ids(root_entity)
                if len(ground_truth_triples) < k:
                    result[root_entity] = ground_truth_triples
                else:
                    result[root_entity] = self.ground_truth_triple_ids(root_entity)[:k]
        return {root_entity: self.ground_truth_triple_ids(root_entity) for root_entity in self.root_entity_ids()}

    def f1_score(self, k: int = None, no_rel: bool = False):
        return WikESSummaryEvaluator(
            super().root_entities(),
            self._ground_truths,
            self._predicted_summaries
        ).evaluate_f1(k, no_rel)

    def map_score(self, k: int = None, no_rel: bool = False):
        return WikESSummaryEvaluator(
            super().root_entities(),
            self._ground_truths,
            self._predicted_summaries
        ).evaluate_map(k, no_rel)



import logging
from collections import defaultdict
from typing import Union, List, Optional, Dict, Tuple

import networkx as nx

from wikes_toolkit.base.graph_components import Entity, RootEntity, Triple, Predicate
from wikes_toolkit.base.versions import DatasetName
from wikes_toolkit.esbm.esbm_graph_components import ESBMBaseGraph, ESBMEntity, ESBMTriple, ESBMPredicate, \
    ESBMRootEntity

logger = logging.getLogger(__name__)


class ESBMGraph(ESBMBaseGraph):
    def __init__(self, G: nx.MultiDiGraph, dataset: DatasetName,
                 root_entity_formatter: Optional[callable] = None,
                 entity_formatter: Optional[callable] = None,
                 predicate_formatter: Optional[callable] = None, triple_formatter: Optional[callable] = None):
        super().__init__(G, dataset, root_entity_formatter, entity_formatter, predicate_formatter, triple_formatter)

    def _extract_gold_summaries(self, data: dict[str, any], triple: ESBMTriple):
        for i in range(6):
            if f"in_gold_top5_{i}" in data:
                order = data[f"in_gold_top5_{i}"]
                self._gold_top_5[data['summary_for']][i].insert(order, triple)
            if f"in_gold_top10_{i}" in data:
                order = data[f"in_gold_top10_{i}"]
                self._gold_top_10[data['summary_for']][i].insert(order, triple)

    def _initialize(self):
        super()._initialize()
        self._entities: Dict[str, Union[ESBMEntity, ESBMRootEntity]] = {}
        self._root_entities: Dict[str, ESBMRootEntity] = {}
        self._predicates: Dict[str, ESBMPredicate] = {}
        self._triples: Dict[Tuple[str, str, str], ESBMTriple] = {}
        self._gold_top_5: Dict[str, List[List[ESBMTriple]]] = defaultdict(lambda: [list() for _ in range(6)])
        self._gold_top_10: Dict[str, List[List[ESBMTriple]]] = defaultdict(lambda: [list() for _ in range(6)])

        logger.debug(f"Initializing ESBMGraph {self._dataset_name}...")

        for node, data in self._G.nodes(data=True):
            if data.get('is_root', False):
                root_entity = ESBMRootEntity(
                    identifier=node,
                    eid=data.get('eid'),
                    category=data.get('category'),
                    str_formatter=self._root_entity_formatter
                )
                self._root_entities[node] = root_entity

            entity = ESBMEntity(
                identifier=node,
                str_formatter=self._entity_formatter
            )
            self._entities[node] = entity
        logger.debug(f"Entities: {len(self._entities)} initialized.")

        # sort root entities based on eid
        self._root_entities = dict(sorted(self._root_entities.items(), key=lambda x: x[1].eid))
        logger.debug(f"Root Entities: {len(self._root_entities)} initialized.")

        logger.debug("Initializing triples...")
        for u, v, data in self._G.edges(data=True):
            predicate_id = data['predicate']
            if predicate_id not in self._predicates:
                self._predicates[predicate_id] = ESBMPredicate(
                    predicate_id=predicate_id,
                    str_formatter=self._predicate_formatter
                )

            triple = ESBMTriple(
                subject_entity=self._entities[u],
                predicate=self._predicates[predicate_id],
                object_entity=self._entities[v],
                str_formatter=self._triple_formatter
            )
            self._triples[(u, predicate_id, v)] = triple

            if 'summary_for' in data:
                self._extract_gold_summaries(data, triple)

    def root_entities(self) -> List[ESBMRootEntity]:
        return super().root_entities()

    def entities(self) -> List[ESBMEntity]:
        return super().entities()

    def triples(self) -> List[ESBMTriple]:
        return super().triples()

    def predicates(self) -> List[ESBMPredicate]:
        return super().predicates()

    def fetch_entity(self, entity: Union[Entity, str]) -> ESBMEntity:
        return super().fetch_entity(entity)

    def fetch_root_entity(self, entity: Union[RootEntity, str]) -> ESBMRootEntity:
        return super().fetch_root_entity(entity)

    def fetch_predicate(self, predicate: Union[Predicate, str]) -> ESBMPredicate:
        return super().fetch_predicate(predicate)

    def fetch_triple(self, triple: Union[
        Triple, Tuple[
            Union[Entity, str],
            Union[Predicate, str],
            Union[Entity, str]
        ],

    ]) -> ESBMTriple:
        return super().fetch_triple(triple)

    def neighbors(self, entity: Union[Entity, str]) -> List[ESBMTriple]:
        return super().neighbors(entity)

    def all_gold_top_k(self, k: int) -> Dict[str, List[List[Tuple[str, str, str]]]]:
        if k not in [5, 10]:
            raise ValueError("k should be 5 or 10")
        gold_top_k = self._gold_top_5 if k == 5 else self._gold_top_10
        result = defaultdict(lambda: [list() for _ in range(6)])
        for root_entity_id, top_list in gold_top_k.items():
            for i, items in enumerate(top_list):
                result[root_entity_id][i] = [
                    (t.subject_entity.identifier, t.predicate.predicate_id, t.object_entity.identifier) for t in items]
        return result

    def gold_top_5(self, root_entity: Union[ESBMRootEntity, str], annotator_index: int) -> List[ESBMTriple]:
        if annotator_index < 0 or annotator_index > 5:
            raise ValueError("Annotator index should be between 0 and 5.")
        return self._gold_top_5[self.fetch_root_entity_id(root_entity)][annotator_index]

    def gold_top_10(self, root_entity: Union[ESBMRootEntity, str], annotator_index: int) -> List[ESBMTriple]:
        if annotator_index < 0 or annotator_index > 5:
            raise ValueError("Annotator index should be between 0 and 5.")
        return self._gold_top_10[self.fetch_root_entity_id(root_entity)][annotator_index]



from wikes_toolkit.base.versions import DatasetName, DatasetVersion


class WikESVersions:

    @staticmethod
    def available_versions():
        subclasses = set()

        def get_all_subclasses(cls):
            for attr_name in dir(cls):
                attr = getattr(cls, attr_name)
                if isinstance(attr, type) and issubclass(attr, DatasetVersion):
                    get_all_subclasses(attr)
                elif isinstance(attr, type) and issubclass(attr, DatasetName):
                    subclasses.add(attr)

        get_all_subclasses(WikESVersions)
        return tuple(subclasses)

    class V1(DatasetVersion):
        version = '1.0.5'
        base_url = f'https://github.com/msorkhpar/wiki-entity-summarization/releases/download/{version}/'

        @staticmethod
        def get_dataset_url(dataset: DatasetName) -> str:
            return WikESVersions.V1.base_url + dataset.value + ".pkl"

        @staticmethod
        def get_version() -> str:
            return WikESVersions.V1.version

        class WikiLitArt(DatasetName):
            SMALL = 'WikiLitArt-s'
            SMALL_TRAIN = 'WikiLitArt-s-train'
            SMALL_VALIDATION = 'WikiLitArt-s-val'
            SMALL_TEST = 'WikiLitArt-s-test'
            MEDIUM = 'WikiLitArt-m'
            MEDIUM_TRAIN = 'WikiLitArt-m-train'
            MEDIUM_VALIDATION = 'WikiLitArt-m-val'
            MEDIUM_TEST = 'WikiLitArt-m-test'
            LARGE = 'WikiLitArt-l'
            LARGE_TRAIN = 'WikiLitArt-l-train'
            LARGE_VALIDATION = 'WikiLitArt-l-val'
            LARGE_TEST = 'WikiLitArt-l-test'

            def get_dataset_url(self):
                return WikESVersions.V1.get_dataset_url(self)

            def get_version(self):
                return WikESVersions.V1.get_version()

        class WikiCinema(DatasetName):
            SMALL = 'WikiCinema-s'
            SMALL_TRAIN = 'WikiCinema-s-train'
            SMALL_VALIDATION = 'WikiCinema-s-val'
            SMALL_TEST = 'WikiCinema-s-test'
            MEDIUM = 'WikiCinema-m'
            MEDIUM_TRAIN = 'WikiCinema-m-train'
            MEDIUM_VALIDATION = 'WikiCinema-m-val'
            MEDIUM_TEST = 'WikiCinema-m-test'
            LARGE = 'WikiCinema-l'
            LARGE_TRAIN = 'WikiCinema-l-train'
            LARGE_VALIDATION = 'WikiCinema-l-val'
            LARGE_TEST = 'WikiCinema-l-test'

            def get_dataset_url(self):
                return WikESVersions.V1.get_dataset_url(self)

            def get_version(self):
                return WikESVersions.V1.get_version()

        class WikiPro(DatasetName):
            SMALL = 'WikiPro-s'
            SMALL_TRAIN = 'WikiPro-s-train'
            SMALL_VALIDATION = 'WikiPro-s-val'
            SMALL_TEST = 'WikiPro-s-test'
            MEDIUM = 'WikiPro-m'
            MEDIUM_TRAIN = 'WikiPro-m-train'
            MEDIUM_VALIDATION = 'WikiPro-m-val'
            MEDIUM_TEST = 'WikiPro-m-test'
            LARGE = 'WikiPro-l'
            LARGE_TRAIN = 'WikiPro-l-train'
            LARGE_VALIDATION = 'WikiPro-l-val'
            LARGE_TEST = 'WikiPro-l-test'

            def get_dataset_url(self):
                return WikESVersions.V1.get_dataset_url(self)

            def get_version(self):
                return WikESVersions.V1.get_version()

        class WikiProFem(DatasetName):
            SMALL = 'WikiProFem-s'
            SMALL_TRAIN = 'WikiProFem-s-train'
            SMALL_VALIDATION = 'WikiProFem-s-val'
            SMALL_TEST = 'WikiProFem-s-test'
            MEDIUM = 'WikiProFem-m'
            MEDIUM_TRAIN = 'WikiProFem-m-train'
            MEDIUM_VALIDATION = 'WikiProFem-m-val'
            MEDIUM_TEST = 'WikiProFem-m-test'
            LARGE = 'WikiProFem-l'
            LARGE_TRAIN = 'WikiProFem-l-train'
            LARGE_VALIDATION = 'WikiProFem-l-val'
            LARGE_TEST = 'WikiProFem-l-test'

            def get_dataset_url(self):
                return WikESVersions.V1.get_dataset_url(self)

            def get_version(self):
                return WikESVersions.V1.get_version()



from typing import List, Tuple, Union


def f1(
        gold_summary: List[Union[Tuple[str, str, str], Tuple[str, str]]],
        algo_summary: List[Union[Tuple[str, str, str], Tuple[str, str]]]
) -> float:
    gold_set = set(gold_summary)
    algo_set = set(algo_summary)

    correct = len(gold_set.intersection(algo_set))
    precision = correct / len(algo_summary)
    recall = correct / len(gold_summary)
    return 2 * precision * recall / (precision + recall) if correct != 0 else 0


def map(gold_summary: List[Union[Tuple[str, str, str], Tuple[str, str]]],
        algo_rank: List[Union[Tuple[str, str, str], Tuple[str, str]]]) -> float:
    relevant_count = 0
    cumulative_precision = 0
    gold_set = set(gold_summary)

    for i, summary in enumerate(algo_rank):
        if summary in gold_set:
            relevant_count += 1
            precision_at_i = relevant_count / (i + 1)
            cumulative_precision += precision_at_i

    return cumulative_precision / len(gold_summary) if relevant_count != 0 else 0


def f1_norel(
        gold_summary: List[Tuple[str, str, str]],
        algo_summary: List[Tuple[str, str, str]]
) -> float:
    gold_set = [(s, o) for s, p, o in gold_summary]
    algo_set = [(s, o) for s, p, o in algo_summary]
    return f1(gold_set, algo_set)


def map_norel(gold_summary: List[Tuple[str, str, str]],
              algo_rank: List[Tuple[str, str, str]]) -> float:
    gold_set = [(s, o) for s, p, o in gold_summary]
    algo_rank = [(s, o) for s, p, o in algo_rank]
    return map(gold_set, algo_rank)



import logging
from typing import Union, Tuple

import networkx as nx
import pandas as pd

from wikes_toolkit.base.versions import DatasetName
from wikes_toolkit.wikes import wikes_exporter
from wikes_toolkit.wikes.wikies_graph_components import WikESBaseGraph

logger = logging.getLogger(__name__)


class PandasWikESGraph(WikESBaseGraph):

    def __init__(self, G: nx.MultiDiGraph, dataset_name: DatasetName):
        super().__init__(G, dataset_name)

    def _initialize(self):
        logger.debug("Initializing PandasWikESGraph...")

        nodes_data = [
            {
                'identifier': node,
                'wikidata_label': data.get('wikidata_label'),
                'wikidata_description': data.get('wikidata_desc'),
                'wikipedia_id': int(data.get('wikipedia_id')) if data.get('wikipedia_id') else None,
                'wikipedia_title': data.get('wikipedia_title'),
                'category': data.get('category', None)
            }
            for node, data in self._G.nodes(data=True) if data.get('is_root', False)
        ]
        self._root_entities = pd.DataFrame(nodes_data, columns=[
            'identifier', 'wikidata_label', 'wikidata_description', 'wikipedia_id', 'wikipedia_title', 'category'
        ]).set_index('identifier')
        logger.debug(f"Root entities: {self._root_entities.shape[0]} initialized.")
        del nodes_data

        nodes_data = [
            {
                'identifier': node,
                'wikidata_label': data.get('wikidata_label'),
                'wikidata_description': data.get('wikidata_desc'),
                'wikipedia_id': int(data.get('wikipedia_id')) if data.get('wikipedia_id') else None,
                'wikipedia_title': data.get('wikipedia_title')
            }
            for node, data in self._G.nodes(data=True)
        ]
        self._entities = pd.DataFrame(nodes_data, columns=[
            'identifier', 'wikidata_label', 'wikidata_description', 'wikipedia_id', 'wikipedia_title'
        ]).set_index('identifier')
        logger.debug(f"Entities: {self._entities.shape[0]} initialized.")

        del nodes_data

        logger.debug("Initializing triples...")
        edges_data = []
        ground_truths_data = []
        predicates_data = dict()

        for u, v, data in self._G.edges(data=True):
            predicate_id = data['predicate']
            edges_data.append({'subject': u, 'predicate': predicate_id, 'object': v})

            if predicate_id not in predicates_data:
                predicates_data[predicate_id] = (data.get('predicate_label'), data.get('predicate_desc'))

            if 'summary_for' in data:
                ground_truths_data.append({
                    'identifier': data['summary_for'],
                    'subject': u,
                    'predicate': predicate_id,
                    'object': v
                })
        predicate_flat = [
            {
                'identifier': k,
                'predicate_label': v[0],
                'predicate_desc': v[1]
            } for k, v in predicates_data.items()
        ]
        self._predicates = pd.DataFrame(
            predicate_flat,
            columns=['identifier', 'predicate_label', 'predicate_desc']
        ).set_index('identifier')
        del predicate_flat, predicates_data

        self._triples = pd.DataFrame(edges_data, columns=['subject', 'predicate', 'object'])
        del edges_data

        self._ground_truths = pd.DataFrame(
            ground_truths_data,
            columns=['identifier', 'subject', 'predicate', 'object']
        ).set_index('identifier')
        del ground_truths_data

        logger.debug(f"Triples: {self._triples.shape[0]} initialized.")

    def root_entities(self) -> pd.DataFrame:
        return super().root_entities()

    def entities(self) -> pd.DataFrame:
        return super().entities()

    def triples(self) -> pd.DataFrame:
        return super().triples()

    def predicates(self) -> pd.DataFrame:
        return super().predicates()

    def fetch_entity(self, entity: [str, pd.Series]) -> pd.Series:
        return super().fetch_entity(entity)

    def fetch_root_entity(self, entity: [str, pd.Series]) -> pd.Series:
        return super().fetch_root_entity(entity)

    def fetch_predicate(self, predicate: [str, pd.Series]) -> pd.Series:
        return super().fetch_predicate(predicate)

    def fetch_triple(self, triple: Union[Tuple[str, str, str], pd.Series]) -> pd.Series:
        return super().fetch_triple(triple)

    def neighbors(self, entity: [str, pd.Series]) -> pd.DataFrame:
        return super().neighbors(entity)

    def export(self, path: str, license_path: str):
        return wikes_exporter.export(
            path,
            license_path,
            self._G,
            self._dataset_name,
            self._entities,
            self._root_entities,
            self._triples,
            self._predicates,
            self._ground_truths
        )
